-------------------------------------------------------------------------------
-- Entity:      apbtmu
-- File:        apbtmu.vhd
-- Author:      Bjorn Forsman
-- Description: TMU with AMBA APB interface (Leon3 peripheral device)
--
-- Revision list
-- Date         By      Changes
-- 2008-04-04   BF      initial design
-- 2008-04-06   BF      added interrupt handling
-- 
--
--
--
--
--
-- ghdl -a --ieee=synopsys --workdir=gnu/work  --work=work -Pgnu -Pgnu/opencores -Pgnu/gaisler -Pgnu/work apbtmu.vhd
-- ghdl -e --ieee=synopsys --workdir=gnu/work  --work=work -Pgnu/grlib apbtmu
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
library grlib;
use grlib.amba.all;                     -- apb_slv_{in,out}_type

package tmupkg is

  constant MAXTIMERS : integer := 32;
  constant TSELBITS  : integer := 5;    -- log2(MAXTIMERS)

  type tmu_in_type is
    record
      dhalt : std_ulogic;               -- halt timers in debug mode
    end record;

  type tmu_out_type is
    record
      apbi       : apb_slv_in_type;         -- to irq controller
      morebudget : std_logic_vector(MAXTIMERS-1 downto 1); -- only for debug
    end record;

  component apbtmu
    generic (
      pindex   : integer := 0;          -- APB device select signal
      paddr    : integer := 0;          -- 12 MSb of APB address
      pmask    : integer := 16#fff#;    -- APB address mask
      pirq     : integer := 0;          -- which APB irq to generate
      ntimers  : integer range 1 to MAXTIMERS := 16;
      tbits    : integer range 1 to 32        := 32;  -- timer bits
      );
    port (
      rst      : in  std_logic;
      clk      : in  std_logic;
      apbi     : in  apb_slv_in_type;
      apbo     : out apb_slv_out_type;
      tmui     : in  tmu_in_type;
      tmuo     : out tmu_out_type
      );
  end component;

end tmupkg;



library ieee;
use ieee.std_logic_1164.all;
library grlib;
use grlib.amba.all;                     -- apb_slv_{in,out}_type
use grlib.stdlib.all;                   -- report_version and "+" fcn. for std_logic_vector!
use grlib.devices.all;                  -- lib/grlib/amba/devices.vhd
                                        -- (configuration constants)
--pragma translate_off
use std.textio.all;
--pragma translate_on
use work.tmupkg.all;

entity apbtmu is
  generic (
    pindex  : integer := 0;              -- APB device select signal
    paddr   : integer := 0;              -- 12-bit MSB APB address
    pmask   : integer := 16#fff#;        -- APB address mask
    pirq    : integer := 0;              -- which APB interrupt to generate
    ntimers : integer range 1 to MAXTIMERS := 16;
    tbits   : integer range 1 to 32        := 32;  -- timer bits
    );
  port (
    rst     : in  std_logic;
    clk     : in  std_logic;
    apbi    : in  apb_slv_in_type;
    apbo    : out apb_slv_out_type;
    tmui    : in  tmu_in_type;
    tmuo    : out tmu_out_type
    );
end;

architecture rtl of apbtmu is

  constant REVISION : integer := 0;

  constant pconfig : apb_config_type := (
    0 => ahb_device_reg(VENDOR_NTNU, NTNU_TMU, 0, REVISION, pirq),
    1 => apb_iobar(paddr, pmask));

  type timer_reg is
    record
      irq    : std_ulogic;                          -- interrupt pulse
      irqen  : std_ulogic;                          -- interrupt enable
      irqpen : std_ulogic;                          -- interrupt pending
      ticks  : std_logic_vector(tbits-1 downto 0);  -- always increment
      count  : std_logic_vector(tbits-1 downto 0);  -- inc when tmr selected
      limit  : std_logic_vector(tbits-1 downto 0);  -- interrupt when cnt>lim
      period : std_logic_vector(tbits-1 downto 0);  -- replenish period
    end record;

  type timer_reg_vector is array (natural range <> ) of timer_reg;

  type registers is
    record
      enable : std_ulogic;                          -- enable TMU
      tsel   : integer range 0 to ntimers-1;        -- timer select
      timers : timer_reg_vector(0 to ntimers-1);
    end record;

  signal r, rin : registers;

begin

  comb : process(rst, r, apbi, tmui)
    variable v          : registers;
    variable readdata   : std_logic_vector(31 downto 0);
    variable tmpirq     : std_logic_vector(NAHBIRQ-1 downto 0);
    variable morebudget : std_logic_vector(ntimers-1 downto 1);
    variable toirqctrl  : apb_slv_in_type; -- tmp var for irq controller
  begin

    v := r;                             -- copy regs to working var.

    -- increment count for the selected timer (when not in debug mode)
    if tmui.dhalt = '0' then 	-- halt timers in debug mode
      if v.enable = '1' then
        v.timers(r.tsel).count := v.timers(r.tsel).count + 1;
      end if;

      -- deferred server model: replenish budget every period
      -- increment ticks, reset count and replenish budgets for irq tmrs
      for i in 1 to ntimers-1 loop
        v.timers(i).ticks   := v.timers(i).ticks + 1;
        if v.timers(i).ticks >= v.timers(i).period then
          v.timers(i).count := (others => '0');
          v.timers(i).ticks := (others => '0');
        end if;
      end loop;

    end if; -- if tmui.dhalt


    -- internal interrupt signal generation
    for i in 0 to ntimers-1 loop
      if (v.timers(i).count = v.timers(i).limit) then
        if (v.timers(i).irqen = '1') then
          v.timers(i).irq    := '1';
          v.timers(i).irqpen := '1';    -- must be reset by software
        end if;
      else
        v.timers(i).irq      := '0';
      end if;
    end loop;

    -- external interrupt lines
    toirqctrl := apbi;                -- default assignment
    for i in 1 to ntimers-1 loop
      if v.timers(i).count < v.timers(i).limit then
        morebudget(i) := '1';
      else
        morebudget(i) := '0';
      end if;

      --into(i) <= inti(i) and not
      toirqctrl.pirq(i) := apbi.pirq(i) and not
                 (not morebudget(i) and v.timers(i).irqen and v.enable);
    end loop;

    -- Task TMU interrupts overrides
    toirqctrl.pirq(pirq) := (v.timers(0).irq and v.timers(0).irqen) and v.enable;


    -- read registers
    readdata := (others => '0');
    case apbi.paddr(8 downto 2) is
      when "0000000" =>                 -- control reg
        readdata(0)                := v.enable;
        readdata(5 downto 1)       := conv_std_logic_vector(ntimers, TSELBITS);
        readdata(11 downto 6)      := conv_std_logic_vector(tbits, 6);
        readdata(12)               := tmui.dhalt;
      when "0000001" =>                 -- timer select reg
        readdata(TSELBITS-1 downto 0) := conv_std_logic_vector(v.tsel, TSELBITS);
      when "0000010" =>                 -- irqpen
        for i in 0 to ntimers-1 loop
          readdata(i) :=  v.timers(i).irqpen;
        end loop;
      when others =>                    -- timer registers
        for i in 0 to ntimers-1 loop
          if conv_integer(apbi.paddr(8 downto 4))-1 = i then
            case apbi.paddr(3 downto 2) is
              when "00"   =>            -- count reg
                readdata(tbits-1 downto 0) := r.timers(i).count;
              when "01"   =>            -- limit reg
                readdata(tbits-1 downto 0) := r.timers(i).limit;
              when "10"   =>            -- period reg
                readdata(tbits-1 downto 0) := r.timers(i).period;
              when "11"   =>            -- timer n control reg
                readdata(0)                := r.timers(i).irqen;
                readdata(1)                := r.timers(i).irqpen;
              when others =>
            end case;
          end if;
        end loop;
    end case;


    -- write registers
    if (apbi.psel(pindex) and apbi.penable and apbi.pwrite) = '1' then
      case apbi.paddr(8 downto 2) is
        when "0000000" =>               -- control reg
          v.enable            := apbi.pwdata(0);
        when "0000001" =>               -- timer select reg
          v.tsel              := conv_integer(apbi.pwdata(TSELBITS-1 downto 0));
        when "0000010" =>               -- irqpen reg
          for i in 0 to ntimers-1 loop
            v.timers(i).irqpen := apbi.pwdata(i);
          end loop;
        when others =>
          for i in 0 to ntimers-1 loop
            if conv_integer(apbi.paddr(8 downto 4))-1 = i then
              case apbi.paddr(3 downto 2) is
                when "00"   =>          -- count reg
                  v.timers(i).count  := apbi.pwdata(tbits-1 downto 0);
                when "01"   =>          -- limit reg
                  v.timers(i).limit  := apbi.pwdata(tbits-1 downto 0);
                when "10"   =>          -- period reg
                  v.timers(i).period := apbi.pwdata(tbits-1 downto 0);
                when "11"   =>          -- timer n control reg
                  v.timers(i).irqen  := apbi.pwdata(0);
                  v.timers(i).irqpen := apbi.pwdata(1);
                when others =>
              end case;
            end if;
          end loop;
      end case;
    end if;


    -- reset
    if rst = '0' then
      for i in 0 to ntimers-1 loop
        v.timers(i).irq    := '0';
        v.timers(i).irqen  := '0';
        v.timers(i).irqpen := '0';
        v.timers(i).ticks  := (others => '0');
        v.timers(i).count  := (others => '0');
        v.timers(i).limit  := (others => '1');
        v.timers(i).period := (others => '1');
      end loop;
      morebudget           := (others => '1'); -- irq timers
      v.enable             := '0';      -- disable counting
      v.tsel               := 0;        -- start with the task timer selected
    end if;


    -- outputs
    rin         <= v;                   -- copy working variable to signal
    apbo.prdata <= readdata;            -- drive apb read bus
    tmuo.apbi   <= toirqctrl;
    tmuo.morebudget(ntimers-1 downto 1) <= morebudget;
  end process;


  -- concurrent assignments
  apbo.pindex  <= pindex;
  apbo.pconfig <= pconfig;


  -- registers
  regs : process(clk)
  begin
    if rising_edge(clk) then
      r <= rin;
    end if;
  end process;


-- boot message

-- pragma translate_off
  bootmsg          : report_version
    generic map ("apbtmu" & tost(pindex) &
                 ": Time Management Unit rev " & tost(REVISION) &
                 ", with " & tost(ntimers) &
                 " " & tost(tbits) & "-bit timers" & ", irq " & tost(pirq));
-- pragma translate_on

end;
