// HAL for the Time Management Unit (TMU) for use with eCos

#ifndef HAL_TMU_H // prevent multiple inclusions
#define HAL_TMU_H

#include <pkgconf/hal_sparc.h>         // CYGPKG_TMUADDR
#include <cyg/hal/hal_io.h>            // HAL IO functions HAL_READ/WRITE_XXX

// protect against build errors when TMU is not enabled for this system
#ifndef CYGPKG_TMUADDR
#define CYGPKG_TMUADDR 0x80001000
#endif

#ifndef TMU_MAXTIMERS
#define TMU_MAXTIMERS 32
#endif

// bit masks
#define TMU_CONFIG_ENABLE_COUNT 0x00000001
#define TIMER_CONTROL_IRQEN     0x00000001
#define TIMER_CONTROL_IRQPEN    0x00000002

struct timer_regs {                    // address offset
  volatile unsigned int count;	       // 0x0
  volatile unsigned int limit;	       // 0x4
  volatile unsigned int period;	       // 0x8
  volatile unsigned int control;       // 0xC
};

struct tmu {                           // address offset
  volatile unsigned int config;        // 0x0
  volatile unsigned int tsel;	       // 0x4 timer select reg
  volatile unsigned int irqpen;	       // 0x8 irqpen for all timers
  volatile unsigned int dummy;	       // 0xC unused
  struct timer_regs timer[TMU_MAXTIMERS];
};

// reg addresses
#define TMU_CONFIG         (&(p->config))
#define TMU_TSEL           (&(p->tsel))
#define TMU_IRQPEN         (&(p->irqpen))
#define TIMER_COUNT(i)     (&(p->timer[i].count))
#define TIMER_LIMIT(i)     (&(p->timer[i].limit))
#define TIMER_PERIOD(i)    (&(p->timer[i].period))
#define TIMER_CONTROL(i)   (&(p->timer[i].control))

//#define SET_BIT(word,bit) (word = word | (1<<bit))
//#define CLR_BIT(word,bit) (word = word & ~(1<<bit))

unsigned int tmu_ntimers(void);
unsigned int tmu_tbits(void);
void tmu_select_tmr(int timer); // tråder har ikke rettigheter...
unsigned int tmu_cur_tmr(void);
void tmu_enable_irq(int timer);
void tmu_disable_irq(int timer);
void tmu_clear_irqpen(void);
unsigned int tmu_read_irqpen(void);

// for application
void tmu_reset(void); // set TMU to the state it enters upon reset
void tmu_enable(void); // enable counters (and in effect, interrupts)
void tmu_disable(void); // freezes timers
void tmu_tmr_setup(int timer, int limit, int period, int irqen); // for interrupt timers

// for use in threads (thread API to TMU)
// must be used when in the threads context
void tmu_task_tmr_setup(int limit, int period, int irqen);
unsigned int tmu_task_tmr_get_count(void);
void tmu_task_tmr_set_count(unsigned int val);
unsigned int tmu_task_tmr_get_limit(void);
void tmu_task_tmr_set_limit(unsigned int val);

void tmu_task_tmr_set_irqen(int irqen); // set/clear
unsigned int tmu_task_tmr_get_irqen(void); // returns 1 or 0
void tmu_task_tmr_set_irqpen(int irqpen); // set/clear
unsigned int tmu_task_tmr_get_irqpen(void); // returns 1 or 0

#endif // !HAL_TMU_H

