// HAL for the Time Management Unit (TMU) for use with eCos

#include <cyg/hal/hal_io.h>         // HAL IO functions HAL_READ/WRITE_XXX
#include <cyg/hal/hal_arch.h>		// CYGPKG_TMUADDR
#include <cyg/hal/hal_tmu.h>

static struct tmu *p = (struct tmu *)CYGPKG_TMUADDR; // base pointer to TMU regs


void tmu_enable(void)
{
  HAL_WRITE_UINT32(TMU_CONFIG, TMU_CONFIG_ENABLE_COUNT);
}

void tmu_disable(void)
{
  HAL_WRITE_UINT32(TMU_CONFIG, 0);
}

unsigned int tmu_ntimers(void)
{
  unsigned int val;
  HAL_READ_UINT32(TMU_CONFIG, val);
  val = ((val)>>1) & 0x1f; // 5 bits: bit1-bit5 inclusive
  return val;
}

unsigned int tmu_tbits(void)
{
  unsigned int val;
  HAL_READ_UINT32(TMU_CONFIG, val);
  val = ((val)>>6) & 0x3f; // 6 bits: bit6-bit11 inclusive
  return val;
}

void tmu_select_tmr(int timer)
{
  HAL_WRITE_UINT32(TMU_TSEL, timer);
}

unsigned int tmu_cur_tmr(void)
{
  int timer;
  HAL_READ_UINT32(TMU_TSEL, timer);
  return timer;
}

void tmu_enable_irq(int timer)
{
  unsigned int val;
  HAL_READ_UINT32(TIMER_CONTROL(timer), val);
  HAL_WRITE_UINT32(TMU_CONFIG, val | TIMER_CONTROL_IRQEN);
}

void tmu_disable_irq(int timer)
{
  unsigned int val;
  HAL_READ_UINT32(TIMER_CONTROL(timer), val);
  HAL_WRITE_UINT32(TMU_CONFIG, val & ~TIMER_CONTROL_IRQEN);
}

void tmu_clear_irqpen(void)
{
  HAL_WRITE_UINT32(TMU_IRQPEN, 0);
}

unsigned int tmu_read_irqpen(void)
{
  unsigned int val;
  HAL_READ_UINT32(TMU_IRQPEN, val);
  return val;
}

void tmu_tmr_setup(int timer, int limit, int period, int irqen)
{
  HAL_WRITE_UINT32(TIMER_COUNT(timer), 0);
  HAL_WRITE_UINT32(TIMER_LIMIT(timer), limit);
  HAL_WRITE_UINT32(TIMER_PERIOD(timer), period);
  if (irqen==0)
    HAL_WRITE_UINT32(TIMER_CONTROL(timer), 0);
  else
    HAL_WRITE_UINT32(TIMER_CONTROL(timer), TIMER_CONTROL_IRQEN);
}

void tmu_reset(void) // set TMU int the state it enters upon reset
{
  int i;
  int ntimers = tmu_ntimers();

  HAL_WRITE_UINT32(TMU_CONFIG, 0);
  HAL_WRITE_UINT32(TMU_TSEL, 0);
  HAL_WRITE_UINT32(TMU_IRQPEN, 0);

  for (i=0; i<ntimers; i++) {
    HAL_WRITE_UINT32(TIMER_COUNT(i), 0);
    HAL_WRITE_UINT32(TIMER_LIMIT(i), 0xffffffff);
    HAL_WRITE_UINT32(TIMER_PERIOD(i), 0xffffffff);
    HAL_WRITE_UINT32(TIMER_CONTROL(i), 0);
  }
}

// Thread API to the TMU
void tmu_task_tmr_setup(int limit, int period, int irqen)
{
  HAL_WRITE_UINT32(TIMER_COUNT(0), 0);
  HAL_WRITE_UINT32(TIMER_LIMIT(0), limit);
  HAL_WRITE_UINT32(TIMER_PERIOD(0), period);
  if (irqen==0)
    HAL_WRITE_UINT32(TIMER_CONTROL(0), 0);
  else
    HAL_WRITE_UINT32(TIMER_CONTROL(0), TIMER_CONTROL_IRQEN);
}

unsigned int tmu_task_tmr_get_count(void)
{
  unsigned int val;
  HAL_READ_UINT32(TIMER_COUNT(0), val);
  return val;
}

void tmu_task_tmr_set_count(unsigned int val)
{
  HAL_WRITE_UINT32(TIMER_COUNT(0), val);
}

unsigned int tmu_task_tmr_get_limit(void)
{
  unsigned int val;
  HAL_READ_UINT32(TIMER_LIMIT(0), val);
  return val;
}

void tmu_task_tmr_set_limit(unsigned int val)
{
  HAL_WRITE_UINT32(TIMER_LIMIT(0), val);
}

void tmu_task_tmr_set_irqen(int irqen)
{
  if (irqen==0)
    HAL_WRITE_UINT32(TIMER_CONTROL(0), 0);
  else
    HAL_WRITE_UINT32(TIMER_CONTROL(0), TIMER_CONTROL_IRQEN);
}

unsigned int tmu_task_tmr_get_irqen(void) // returns 1 or 0
{
  unsigned int val;
  HAL_READ_UINT32(TIMER_CONTROL(0), val);
  return (val & TIMER_CONTROL_IRQEN);
}

void tmu_task_tmr_set_irqpen(int irqpen) // set/clear
{
  unsigned int val;
  HAL_READ_UINT32(TIMER_CONTROL(0), val);
  if (irqpen==0)
    HAL_WRITE_UINT32(TIMER_CONTROL(0), val & ~TIMER_CONTROL_IRQPEN);
  else
    HAL_WRITE_UINT32(TIMER_CONTROL(0), val | TIMER_CONTROL_IRQPEN);
}

unsigned int tmu_task_tmr_get_irqpen(void) // returns 1 or 0
{
  unsigned int val;
  HAL_READ_UINT32(TIMER_CONTROL(0), val);
  return (val & TIMER_CONTROL_IRQPEN)>>1;
}

