///////////////////////////////////////////////////////////////////////////////
// Testing the IRQ timers of the Time Management Unit (TMU).
//
// This program utilizes the IRQ timers of the TMU to limit the interarrival
// times of an aperiodic task. (Aperiodic tasks have no lower bound on their
// inter arrival times, sporadic tasks do.) The test consists of two threads,
// one thread generating interrupts (with the help of the GPIO unit) at
// variable intervals and the other thread is released by this interrupt,
// or more correctly, by the corresponding DSR.
// The TMU is setup to limit GPIO interrupts occuring too frequent.
// This makes the aperiodic thread sporadic. The sporadic thread has the
// highest priority, thus it will always preempt the other thread
// (when not blocked).
//
///////////////////////////////////////////////////////////////////////////////

#include <cyg/kernel/kapi.h>           // All the kernel specific stuff
#include <stdio.h>                     // printf()
#include <cyg/hal/hal_arch.h>          // CYGNUM_HAL_STACK_SIZE_TYPICAL/MINIMUM
#include <cyg/hal/hal_tmu.h>           // TMU header

#define NTHREADS 2
#define STACKSIZE ( CYGNUM_HAL_STACK_SIZE_MINIMUM )	// + 4096

#define GPIO_BASE          0x80000800
#define GPIO_PORT_IN       ( GPIO_BASE )
#define GPIO_PORT_OUT      ( GPIO_BASE + 0x4 )
// a '1' defines a port bit as output
#define GPIO_PORT_DIR      ( GPIO_BASE + 0x8 )
// a '1' enables irq for this line (the VHDL generic imask must also be set)
#define GPIO_PORT_IRQ_MASK ( GPIO_BASE + 0xC )
// a '1' indicates active high level irq or positive edge triggered irq
#define GPIO_PORT_IRQ_POL  ( GPIO_BASE + 0x10 )
// a '1' indicates edge triggered irq
#define GPIO_PORT_IRQ_EDGE ( GPIO_BASE + 0x14 )

#define LED0_MASK 1                     // for the sporadic task to indicate its release
#define LED1_MASK 2
#define LED2_MASK 4
#define LED3_MASK 8                     // indicates when an irq is attempted
#define GPIO_IRQ CYGNUM_HAL_INTERRUPT_3 // irq 3 is LED3

// New datatypes
struct thread_arg_t {
  int message;
};

// Function prototypes
void thread1(cyg_addrword_t data);
void thread2(cyg_addrword_t data);
cyg_uint32 isr(cyg_vector_t vector, cyg_addrword_t data);
void dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);

// Global variables
static cyg_thread thread_obj[NTHREADS];	 // space for thread objects
static cyg_handle_t thread[NTHREADS];    // handles for the threads
static char stack[NTHREADS][STACKSIZE];	 // space for n threads stacks
static struct thread_arg_t thread_arg[NTHREADS];
cyg_handle_t isr_handle;                 // a handle for the ISR
cyg_interrupt intr;                      // for holding the interrupt handler and associated data.
cyg_sem_t sem;                           // aperiodic thread blocks on this
cyg_mutex_t lock;                        // GPIO



// This is the application entry point
void cyg_user_start(void)
{
  int i;

  printf("Entering cyg_user_start() function\n");

  // TMU
  tmu_reset();
  tmu_enable();

  int period = 40000000 * 2; // sys_clk * wanted_period_in_seconds
  // 1000 instruksjoner ~ 0x10000 sykler (65536)
  // tok tiden en isr bruker (kan variere): 0x629bc
  tmu_tmr_setup(GPIO_IRQ, 0x1, period, 1);	// setup for aperiodic thread (or its irq)

  cyg_semaphore_init(&sem, 0); // initial value 0 thread will start blocked
  cyg_mutex_init(&lock);       // for GPIO port mutual exclusion

  // GPIO
  HAL_WRITE_UINT32(GPIO_PORT_OUT, 0);
  HAL_WRITE_UINT32(GPIO_PORT_DIR, 0x1ff);       // set leds to output, others to input
  HAL_WRITE_UINT32(GPIO_PORT_IRQ_EDGE, 0xffff); // edge triggered
  HAL_WRITE_UINT32(GPIO_PORT_IRQ_POL, 0xffff);  // positive edge triggered
  HAL_WRITE_UINT32(GPIO_PORT_IRQ_MASK, LED3_MASK); // VHDL imask is now 0xffff

  // register ISRs
  cyg_interrupt_create(GPIO_IRQ, 0, 0, isr, dsr, &isr_handle, &intr);
  cyg_interrupt_attach(isr_handle);
  cyg_interrupt_unmask(GPIO_IRQ);

  // thread arguments
  thread_arg[0].message = 1;
  thread_arg[1].message = 2;

  cyg_thread_create(4, thread1, (cyg_addrword_t) &thread_arg[0], "Thread 1",
                   (void *) stack[0], STACKSIZE, &thread[0], &thread_obj[0]);
  cyg_thread_create(3, thread2, (cyg_addrword_t) &thread_arg[1], "Thread 2",
                   (void *) stack[1], STACKSIZE, &thread[1], &thread_obj[1]);

  for (i=0; i<NTHREADS; i++) {
    cyg_thread_resume(thread[i]);
  }
}



// Interrupt service routine
cyg_uint32 isr(cyg_vector_t vector, cyg_addrword_t data)
{
  cyg_bool_t dsr_required = 1;
//  printf("  --> ISR vector=%i, data=0x%x\n", vector, data); // causes deadlock

  return dsr_required ? CYG_ISR_CALL_DSR : CYG_ISR_HANDLED;
}



// Deferred service routine
void dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
  //printf("  --> DSR vector=%i, count=%d, data=0x%x\n", vector, count, data); // causes deadlock
  cyg_semaphore_post(&sem); // release aperiodic thread
}



// this thread is released by a GPIO IRQ, when the DSR is run
void thread2(cyg_addrword_t data)
{
  struct thread_arg_t *arg = (struct thread_arg_t *) data;

  printf("Beginning execution; thread %d\n", arg->message);

  for (;;) {
    cyg_semaphore_wait(&sem); // block until released in DSR
    printf("  --> Thread %d: released!\n", arg->message);
    cyg_mutex_lock(&lock);
    HAL_WRITE_UINT32(GPIO_PORT_OUT, LED0_MASK);
    cyg_thread_delay(2); // don't turn of LED immediately
    HAL_WRITE_UINT32(GPIO_PORT_OUT, 0);
    cyg_mutex_unlock(&lock);
  }
}



// this thread generates an IRQ via the GPIO
void thread1(cyg_addrword_t data)
{
  struct thread_arg_t *arg = (struct thread_arg_t *) data;
  int delay;
  int min = 5;
  int max = 500;
  //int step=5;                        // triangular and sawtooth wave
  //int dir=1;                         // for triangular wave
  unsigned int t1, t2, actual_delay; // debug

  printf("Beginning execution; thread %d\n", arg->message);
  delay = min;

  for (;;) {
    tmu_task_tmr_set_count(0);
    // lit LED and generate IRQ
    cyg_mutex_lock(&lock);
    HAL_WRITE_UINT32(GPIO_PORT_OUT, LED3_MASK);
    // don't turn of LED immediately
    cyg_thread_delay(2);
    HAL_WRITE_UINT32(GPIO_PORT_OUT, 0);
    cyg_mutex_unlock(&lock);
    printf("Thread %d: generated irq and now a delay of %d clock ticks\n", arg->message, delay);
    t1 = cyg_current_time();
    cyg_thread_delay(delay);
    t2 = cyg_current_time();
    actual_delay = t2-t1;
    if ( actual_delay != delay )
      printf("  --> Thread %d: actual delay %d clock ticks\n", arg->message, actual_delay);

    delay *= 1.2;                 // exponential intervals, better
    if ( delay >= max ) delay = min;
/*    dir = (delay <= min)? 1:dir; // triangle wave intervals
    dir = (delay >= max)? -1:dir;
    delay += step*dir;
*/
/*    delay += step;                 // sawtooth wave intervals
    if ( delay >= max ) delay = min;
*/
  }
}

