///////////////////////////////////////////////////////////////////////////////
// Testing the Task Timer of the Time Management Unit (TMU).
//
// This demonstrates the use of the Task Timer of the TMU, having two running
// threads. The TMU generates an exception when the current running thread
// exceeds its budget. If a thread does not catch this exception, no action
// is taken (eCos default handler called) and the thread continues.
// In this example, both thread registers an exception handler which jumps
// back to the thread with the help of a trampoline function and setjmp/longjmp.
// Each thread does some random amount of work, and thus the timeout handler
// will be called from time to time.
//
// NOTE: 1. eCos should be configured with thread safe I/O streams.
//          (else printf~s are not protected with mutex.)
//       2. IMPORTANT! Never use a mutex in a place that might get interrupted
//          by the TMU. Because when resuming, locking an already owned mutex
//          fails. This is only an effect of the setjmp/longjmp. If ordinary
//          exceptions (resumption model) are used, there should be no problem.
//       3. eCos exception handlers have been modified to allow setting a
//          return value which will be jumped to if above 10.
///////////////////////////////////////////////////////////////////////////////

#include <cyg/kernel/kapi.h>          // All the kernel specific stuff
#include <stdio.h>                    // printf()
#include <cyg/hal/hal_arch.h>         // CYGNUM_HAL_STACK_SIZE_TYPICAL/MINIMUM
                                      // and CYG_HAL_TMU_IRQ
#include <cyg/hal/hal_tmu.h>          // TMU API header
#include <setjmp.h>

#define NTHREADS 2
#define STACKSIZE ( CYGNUM_HAL_STACK_SIZE_MINIMUM )	// + 4096

#define GPIO_BASE          0x80000800
#define GPIO_PORT_IN       ( GPIO_BASE )
#define GPIO_PORT_OUT      ( GPIO_BASE + 0x4 )
// a '1' defines a port bit as output
#define GPIO_PORT_DIR      ( GPIO_BASE + 0x8 )
// a '1' enables irq for this line (the VHDL generic imask must also be set)
#define GPIO_PORT_IRQ_MASK ( GPIO_BASE + 0xC )
// a '1' indicates active high level irq or positive edge triggered irq
#define GPIO_PORT_IRQ_POL  ( GPIO_BASE + 0x10 )
// a '1' indicates edge triggered irq
#define GPIO_PORT_IRQ_EDGE ( GPIO_BASE + 0x14 )

#define LED0_MASK 1
#define LED1_MASK 2
#define LED2_MASK 4
#define LED3_MASK 8

// TMU timer datatype
struct tmu_t {
  unsigned int count;
  unsigned int limit;
  unsigned int period; // ignored by task timer
  unsigned int irqen;
};

struct thread_arg_t {
  struct tmu_t tmu; // task timer setup args
  int message;
  int a;
  int b;
  jmp_buf env;      // setjmp/longjmp
  cyg_flag_t flag;  // event flag
};

// Function prototypes
void worker(cyg_addrword_t data);
unsigned int tmu_xsr(cyg_addrword_t data, cyg_code_t exception_number, cyg_addrword_t info);
static void jump(void);
void show_threads(void);
void alarm_handler(cyg_handle_t alarmh, cyg_addrword_t data);

// Global variables
static cyg_thread thread_obj[NTHREADS];	 // space for thread objects
static cyg_handle_t thread[NTHREADS];    // handles for the threads
static char stack[NTHREADS][STACKSIZE];	 // space for n threads stacks
static struct thread_arg_t thread_arg[NTHREADS];
cyg_mutex_t lock;                        // GPIO
// alarm stuff
cyg_handle_t counterh;       // a counter handle
cyg_handle_t system_clockh;  // the system clock handle
cyg_handle_t alarmh;         // an alarm handle
cyg_alarm alarm_obj;             // space for alarm objects



// This is the application entry point
void cyg_user_start(void)
{
  int i;

  printf("Entering cyg_user_start() function\n");
  //show_threads();

  #ifdef CYGPKG_HAL_TMU_IRQ
    cyg_interrupt_unmask(CYGPKG_HAL_TMU_IRQ); // handled as an exception
  #endif

  cyg_mutex_init(&lock);       // for GPIO port mutual exclusion

  // GPIO
  HAL_WRITE_UINT32(GPIO_PORT_OUT, 0);
  HAL_WRITE_UINT32(GPIO_PORT_DIR, 0x1ff);       // set leds to output, others to input

  // TMU
  tmu_reset();
  tmu_enable();

  // thread arguments
  thread_arg[0].message = 1;
  thread_arg[0].a = LED0_MASK;
  thread_arg[0].tmu.limit = 1500000;	 // work ~ 200
  thread_arg[0].tmu.period = 0x11111111; // does not matter for task timer
  thread_arg[0].tmu.irqen = 1;
  cyg_flag_init(&(thread_arg[0].flag));  // event flag

  thread_arg[1].message = 2;
  thread_arg[1].a = LED1_MASK;
  thread_arg[1].tmu.limit = 1500000;
  thread_arg[1].tmu.period = 0x22222222; // does not matter for task timer
  thread_arg[1].tmu.irqen = 1;
  cyg_flag_init(&(thread_arg[1].flag));  // event flag

  thread_arg[2].message = 3;
  thread_arg[2].a = LED2_MASK;
  thread_arg[2].tmu.limit = 1500000;
  thread_arg[2].tmu.period = 0x33333333; // does not matter for task timer
  thread_arg[2].tmu.irqen = 1;
  cyg_flag_init(&(thread_arg[2].flag));  // event flag

  // setup periodic alarm
  system_clockh = cyg_real_time_clock();
  cyg_clock_to_counter(system_clockh, &counterh); // make a counter handle from a clock handle
  cyg_alarm_create(counterh, alarm_handler,       // create the alarm
		   (cyg_addrword_t) NULL,
		   &alarmh, &alarm_obj);
  cyg_alarm_initialize(alarmh, cyg_current_time()+100, 25);

  cyg_thread_create(2, worker, (cyg_addrword_t) &thread_arg[0], "Thread 1",
                   (void *) stack[0], STACKSIZE, &thread[0], &thread_obj[0]);
  cyg_thread_create(3, worker, (cyg_addrword_t) &thread_arg[1], "Thread 2",
                   (void *) stack[1], STACKSIZE, &thread[1], &thread_obj[1]);
//  cyg_thread_create(4, worker, (cyg_addrword_t) &thread_arg[2], "Thread 3",
//                   (void *) stack[2], STACKSIZE, &thread[2], &thread_obj[2]);

  for (i=0; i<NTHREADS; i++) {
    cyg_thread_resume(thread[i]);
  }
  show_threads();
}



// Exception Service Routine. Thread context (may block)
unsigned int tmu_xsr(cyg_addrword_t data, cyg_code_t exception, cyg_addrword_t info)
{
  //printf("\n  XSR: data=0x%x exception=%d info=0x%x\n", data, exception, info);
  //cyg_scheduler_lock(); // protect against scheduling when manipulating TMU
  //tmu_task_tmr_set_irqpen(0);
  //cyg_scheduler_unlock();
//  cyg_thread_delay(10000000);
  //return 0;
  return (unsigned int)jump;
}


// This function is not "called" as a ordinary fcn,
// it acts just as a place for XSRs to return to and then longjmp to the
// "real" exception handler inside the thread.
static void
jump(void)
{
  int i, th;

//  printf("\n    jump: th=0x%x   -----\n\n", th);
  th = cyg_thread_self();
  // find the thread with this handle
  for (i=0; i<NTHREADS; i++) {
    if (thread[i] == th) {
      longjmp(thread_arg[i].env, 3);
    }
  }
  printf("\n  ERROR: could not find a thread with th=0x%x. Quitting this thread.\n", th);
  cyg_thread_exit();
}



// invoked as an alarm handler (DSR context)
void alarm_handler(cyg_handle_t alarmh, cyg_addrword_t data)
{
  int i;
  //printf("alarm\n"); // potential deadlock
  for (i=0; i<NTHREADS; i++) {
    cyg_flag_setbits(&(thread_arg[i].flag), 1);
  }
}



// variable execution time
void worker(cyg_addrword_t data)
{
  struct thread_arg_t *arg = (struct thread_arg_t *) data;
  volatile int i = 0;
  volatile int j = 0;
  volatile int work = 50; // the amount of work this thread will attempt to do
  int c; // for reading TMU count

  printf("Beginning execution; thread %d\n", arg->message);
  cyg_exception_set_handler(CYGNUM_HAL_EXCEPTION_OTHERS, (void*)tmu_xsr, 0, NULL, NULL);

  if ( setjmp(arg->env) != 0 ) {
    c = tmu_task_tmr_get_count();
    printf("  --> Thread %d timeout: TMU count=%u work=%d i=%d j=%d\n",
        arg->message, c, work, i, j);
    // do clean up:
    // do I own any mutexes? yes, unlock
    // -eCos have no way of telling who owns a mutex to users
  } else { // after setjmp initializes env (only once)
    tmu_task_tmr_setup(arg->tmu.limit, arg->tmu.period, arg->tmu.irqen);
    //tmu_task_tmr_setup(0xffffffff, arg->tmu.period, arg->tmu.irqen);
    //tmu_task_tmr_setup(1, arg->tmu.period, arg->tmu.irqen);
    //tmu_task_tmr_setup(0x10000, arg->tmu.period, arg->tmu.irqen);
  }

  for (;;) {
    tmu_task_tmr_set_count(0);
    // block on event flag and clear bit when event occurs
    cyg_flag_wait(&(arg->flag), 1, CYG_FLAG_WAITMODE_AND | CYG_FLAG_WAITMODE_CLR);

    //printf("Thread %d: doing %d work\n", arg->message, work); // dangerous!

    // blink a led to indicate thread release
/*    cyg_mutex_lock(&lock);
    HAL_WRITE_UINT32(GPIO_PORT_OUT, arg->a);
    cyg_thread_delay(2);                // don't turn of LED immediately
    HAL_WRITE_UINT32(GPIO_PORT_OUT, 0);
    cyg_mutex_unlock(&lock);
*/
    work = (work>=200) ? 100 : work+50;
    for (i=0; i<work; i++) {
      for (j=0; j<100; j++);
    }

    // this is debug
    c = tmu_task_tmr_get_count();
    //cyg_scheduler_lock();
    tmu_task_tmr_set_irqen(0); // prevents TMU irq while in printf - IMPORTANT!
    //cyg_scheduler_unlock();
    printf("Thread %d: used %7u TMU cycles work=%d\n", arg->message, c, work);
/*    printf("Thread %d: used %u TMU cycles\n", arg->message, c);
    printf("Thread %d: used %u TMU cycles\n", arg->message, c);
    printf("Thread %d: used %u TMU cycles\n", arg->message, c);
    printf("Thread %d: used %u TMU cycles\n", arg->message, c);
*/
    //cyg_scheduler_lock();
    tmu_task_tmr_set_irqen(1);
    //cyg_scheduler_unlock();
  }
}



// debug thing
void show_threads(void)
{
  cyg_handle_t thread = 0;
  cyg_uint16 id = 0;
  cyg_thread_info info;

  printf("  --> Current thread is handle: 0x%x\n", cyg_thread_self());

  while( cyg_thread_get_next( &thread, &id ) ) {

    if( !cyg_thread_get_info( thread, id, &info ) )
      break;

    printf("ID: %04x name: %15s pri: %2d state: %d handle: 0x%x\n",
           info.id, info.name?info.name:"----", info.set_pri, info.state, info.handle);
  }
}

