///////////////////////////////////////////////////////////////////////////////
// Testing the Time Management Unit (TMU) with aperiodic tasks.
// Using gpio unit to genereate irq upon a button push and release a task.
// The TMU prevents the irq to occur too frequently.
//
// NOTE: Program does nothing until BUTTON C is pushed.
//
///////////////////////////////////////////////////////////////////////////////

#include <cyg/kernel/kapi.h>           // All the kernel specific stuff
#include <stdio.h>                     // printf()
#include <stdlib.h>                    // rand()
#include <cyg/hal/hal_arch.h>          // CYGNUM_HAL_STACK_SIZE_TYPICAL/MINIMUM
#include <cyg/hal/hal_tmu.h>           // TMU header

#define NTHREADS 2
#define STACKSIZE ( CYGNUM_HAL_STACK_SIZE_MINIMUM )	// + 4096

#define GPIO_BASE          0x80000800
#define GPIO_PORT_IN       ( GPIO_BASE )
#define GPIO_PORT_OUT      ( GPIO_BASE + 0x4 )
#define GPIO_PORT_DIR      ( GPIO_BASE + 0x8 )
#define GPIO_PORT_IRQ_MASK ( GPIO_BASE + 0xC )		// a '1' enables irq for this line (the VHDL generic imask must also be set)
#define GPIO_PORT_IRQ_POL  ( GPIO_BASE + 0x10 )		// a '1' indicates active high irq or positive edge triggered irq
#define GPIO_PORT_IRQ_EDGE ( GPIO_BASE + 0x14 )		// a '1' indicates edge triggered irq

#define BUTTON_C_MASK 0x0200 // see leon3mp.ucf. button C is bit9 of the GPIO
#define BUTTON_W_MASK 0x0400 // bit10 of the GPIO
#define BUTTON_S_MASK 0x0800 // bit11 of the GPIO
#define BUTTON_E_MASK 0x1000 // bit12 of the GPIO
#define BUTTON_N_MASK 0x2000 // bit13 of the GPIO
#define BUTTONS_MASK (BUTTON_C_MASK | BUTTON_W_MASK | BUTTON_S_MASK | BUTTON_E_MASK | BUTTON_N_MASK)

// TMU timer datatype
struct tmu_t {
  unsigned int count;
  unsigned int limit;
  unsigned int period; // ignored by task timer
  unsigned int irqen;
};

struct thread_arg_t {
  struct tmu_t tmu;
  int message;
  int a;  // work = a + (rand() % b)
  int b;  // work = a + (rand() % b)
};

// Function prototypes
void simple_prog(cyg_addrword_t data);
void aperiodic(cyg_addrword_t data);
cyg_uint32 isr(cyg_vector_t vector, cyg_addrword_t data);
void dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);
void show_threads(void);

// Global variables
static cyg_thread thread_obj[NTHREADS];	 // space for thread objects
static cyg_handle_t thread[NTHREADS];    // handles for the threads
static char stack[NTHREADS][STACKSIZE];	 // space for n threads stacks
static struct thread_arg_t thread_arg[NTHREADS];
cyg_handle_t isr_handle[16];             // a handle for the ISR
cyg_interrupt intr[16];                  // for holding the interrupt handler and associated data.
cyg_sem_t sem;

// other (global) debug vars
int numirq;



// This is the application entry point
void cyg_user_start(void)
{
  int i;
  cyg_priority_t priority=0;
  cyg_addrword_t data=0;

  printf("Entering cyg_user_start() function\n");
  show_threads();

  tmu_reset();
  tmu_enable();
  //tmu_tmr_setup(4, 0x1, 0xff00000, 1);
  cyg_semaphore_init(&sem, 0); // initial value zero so a wait() on this semaphore will block

  // setup GPIO irq
  HAL_WRITE_UINT32(GPIO_PORT_OUT, 0);
  HAL_WRITE_UINT32(GPIO_PORT_IRQ_EDGE, 0xffff); // edge triggered
  HAL_WRITE_UINT32(GPIO_PORT_IRQ_POL, 0xffff); // positive edge triggered
  HAL_WRITE_UINT32(GPIO_PORT_IRQ_MASK, BUTTONS_MASK); // VHDL imask is now 0xffff

  // set leds
  HAL_WRITE_UINT32(GPIO_PORT_DIR, 0x1ff); // set leds to output, others to input

  // register ISRs
//  for (i=1; i<16; i++) {
  for (i=4; i<14; i++) {
    cyg_interrupt_create(i, priority, data, isr, dsr, &isr_handle[i], &intr[i]);
    cyg_interrupt_attach(isr_handle[i]);
    cyg_interrupt_unmask(i);
  }

/*
  cyg_interrupt_create(CYGNUM_HAL_INTERRUPT_4, priority, data, isr, dsr, &isr_handle[0], &intr[0]);
  cyg_interrupt_attach(isr_handle[0]);
  cyg_interrupt_unmask(CYGNUM_HAL_INTERRUPT_4);
*/
  // setup threads
  thread_arg[0].message = 0;
  thread_arg[0].a = 500;
  thread_arg[0].b = 500;
  thread_arg[0].tmu.limit = 60898343;
  thread_arg[0].tmu.period = 0xf0f0f0f0; // does not matter for task timer
  thread_arg[0].tmu.irqen = 0;

  thread_arg[1].message = 1;
  thread_arg[1].a = 1000;
  thread_arg[1].b = 1000;
  thread_arg[1].tmu.limit = 130927353;
  thread_arg[1].tmu.period = 0xf0f0f0f0; // does not matter for task timer
  thread_arg[1].tmu.irqen = 0;

  cyg_thread_create(4, simple_prog, (cyg_addrword_t) &thread_arg[0], "Thread A",
                   (void *) stack[0], STACKSIZE, &thread[0], &thread_obj[0]);
  cyg_thread_create(3, aperiodic, (cyg_addrword_t) &thread_arg[1], "Aperiodic thread",
                   (void *) stack[1], STACKSIZE, &thread[1], &thread_obj[1]);

  for (i=0; i<NTHREADS; i++) {
    cyg_thread_resume(thread[i]);
  }
  show_threads();
}



// Interrupt service routine (should not use syscalls)
cyg_uint32 isr(cyg_vector_t vector, cyg_addrword_t data)
{
  cyg_bool_t dsr_required = 1;

  numirq++;
//  printf("  --> ISR vector=%i, data=0x%x\n", vector, data);

  return dsr_required ? CYG_ISR_CALL_DSR : CYG_ISR_HANDLED;
}



// Deferred service routine (should never use blocking syscalls)
void dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
  // printf (should) use mutex and is not safe here!! Expect deadlock...
  //printf("  --> DSR vector=%i, count=%d, data=0x%x\n", vector, count, data);
  cyg_semaphore_post(&sem); // release thread
}



void aperiodic(cyg_addrword_t data)
{
  struct thread_arg_t *arg = (struct thread_arg_t *) data;
  unsigned int c, i, j;
  unsigned int work;

  printf("Beginning execution; thread data is %d\n", arg->message);
  tmu_task_tmr_setup(arg->tmu.limit, arg->tmu.period, arg->tmu.irqen);

  for (;;) {
    cyg_semaphore_wait(&sem); // block until released in DSR
    tmu_task_tmr_set_count(0);
    //work = arg->a + (rand() % arg->b); // variable amount of work
    printf("  --> Thread %d (aperiodic): released!\n", arg->message);
/*
    for (i=0; i<work; i++) {
      for (j=0; j<1000; j++);
    }
*/
    c = tmu_task_tmr_get_count();
/*    tmu_task_tmr_set_count(0); // prevents interrupts occurring while doing debug-stuff
    printf("Thread %d: used %u TMU cycles\n", message, c);
*/
  }
}



// this is a simple program which runs in a thread
void simple_prog(cyg_addrword_t data)
{
  struct thread_arg_t *arg = (struct thread_arg_t *) data;
  int delay;
  unsigned int c, val;
  int count=0;

  printf("Beginning execution; thread data is %d\n", arg->message);
  tmu_task_tmr_setup(arg->tmu.limit, arg->tmu.period, arg->tmu.irqen);

  for (;;) {
    tmu_task_tmr_set_count(0);
    delay = 10;// + (rand() % 50);
//    printf("Thread %d: and now a delay of %d clock ticks\n", arg->message, delay);

    // binary counter leds
    count = (count + 1) % 0x1ff;
    HAL_WRITE_UINT32(GPIO_PORT_OUT, count);

    HAL_READ_UINT32(GPIO_PORT_IN, val);
//    printf("Thread %d: gpio val=%d\n", arg->message, val);
//    if ( val & BUTTON_C_MASK )
//      printf("Button C pushed\n");

    cyg_thread_delay(delay);
    c = tmu_task_tmr_get_count();
//    tmu_task_tmr_set_count(0); // prevents interrupts occurring while doing debug-stuff
//    printf("Thread %d: used %u TMU cycles\n", message, c);
  }
}



void show_threads(void)
{
  cyg_handle_t thread = 0;
  cyg_uint16 id = 0;
  cyg_thread_info info;

  printf("  --> Current thread is handle: 0x%x\n", cyg_thread_self());

  while( cyg_thread_get_next( &thread, &id ) ) {

    if( !cyg_thread_get_info( thread, id, &info ) )
      break;

    printf("ID: %04x name: %15s pri: %2d state: %d handle: 0x%x\n",
           info.id, info.name?info.name:"----", info.set_pri, info.state, info.handle);
  }
}

