///////////////////////////////////////////////////////////////////////////////
// Testing the Time Management Unit (TMU) with signal handler.
// eCos must be configured with CYGSEM_LIBC_SIGNALS_HWEXCEPTIONS.
// This forces the configuration of global exception handlers. The system
// then defines an exception handler which calls the user supplied signal
// handler (per thread option).
//
// NOTE: configure eCos to generate signals upon hw exceptions
//         CYGSEM_LIBC_SIGNALS_HWEXCEPTIONS.
//       TMU signal is SIGUSR1.
///////////////////////////////////////////////////////////////////////////////

#include <cyg/kernel/kapi.h>           // All the kernel specific stuff
#include <stdio.h>                     // printf()
#include <signal.h>                    // signal()
#include <cyg/hal/hal_arch.h>          // CYGNUM_HAL_STACK_SIZE_TYPICAL/MINIMUM
#include <cyg/hal/hal_tmu.h>           // TMU header

#define NTHREADS 2
#define STACKSIZE ( CYGNUM_HAL_STACK_SIZE_MINIMUM )	// + 4096

// TMU timer datatype
struct tmu_t {
  unsigned int count;
  unsigned int limit;
  unsigned int period; // ignored by task timer
  unsigned int irqen;
};

struct thread_arg_t {
  struct tmu_t tmu;
  int message;
};

// Function prototypes
void thread1(cyg_addrword_t data);
static void sighandler(int signum);
void show_threads(void); // debug

// Global variables
static cyg_thread thread_obj[NTHREADS];	 // space for thread objects
static cyg_handle_t thread[NTHREADS];    // handles for the threads
static char stack[NTHREADS][STACKSIZE];	 // space for n threads stacks
static struct thread_arg_t thread_arg[NTHREADS];
int numsig; // debug



// This is the application entry point
void cyg_user_start(void)
{
  int i;

  printf("Entering cyg_user_start() function\n");
  show_threads();
  #ifdef CYGPKG_HAL_TMU_IRQ
    cyg_interrupt_unmask(CYGPKG_HAL_TMU_IRQ); // handled as an exception
  #endif

  tmu_reset();
  tmu_enable();

  // thread arguments
  thread_arg[0].message = 1;
  thread_arg[0].tmu.limit = 1500000;	 // work ~ 200
  thread_arg[0].tmu.period = 0x11111111; // does not matter for task timer
  thread_arg[0].tmu.irqen = 1;

  thread_arg[1].message = 2;
  thread_arg[1].tmu.limit = 1500000;
  thread_arg[1].tmu.period = 0x22222222; // does not matter for task timer
  thread_arg[1].tmu.irqen = 1;

  cyg_thread_create(3, thread1, (cyg_addrword_t) &thread_arg[0], "Thread 1",
                   (void *) stack[0], STACKSIZE, &thread[0], &thread_obj[0]);
  cyg_thread_create(4, thread1, (cyg_addrword_t) &thread_arg[1], "Thread 2",
                   (void *) stack[1], STACKSIZE, &thread[1], &thread_obj[1]);

  for (i=0; i<NTHREADS; i++) {
    cyg_thread_resume(thread[i]);
  }
  show_threads();
}



//__sighandler_t
static void
sighandler(int signum)
{
  int i;
  int th;

  signal(SIGUSR1, (__sighandler_t)sighandler); // reinstall signal handler
  numsig++;
  th = cyg_thread_self();
  // find the thread with this handle
  for (i=0; i<NTHREADS; i++) {
    if (thread[i] == th) {
      printf("  --> Thread %d: signum %d\n", thread_arg[i].message, signum); // potential locking of already owned mutexes - BAD!
      break;
    }
  }

}



// variable execution time 1
void thread1(cyg_addrword_t data)
{
  struct thread_arg_t *arg = (struct thread_arg_t *) data;
  volatile int i = 0;
  volatile int j = 0;
  volatile int work = 50; // the amount of work this thread will attempt to do
  int c; // for reading TMU count
  int delay;  // ~thread period

  printf("Beginning execution; thread %d\n", arg->message);
  tmu_task_tmr_setup(arg->tmu.limit, arg->tmu.period, arg->tmu.irqen);
  signal(SIGUSR1, (__sighandler_t)sighandler); // register signal handler

  for (;;) {
    tmu_task_tmr_set_count(0);
    delay = 50;
//    printf("Thread %d: doing %d work\n", arg->message, work);
    work = (work>=200) ? 100 : work+50;
    for (i=0; i<work; i++) {
      for (j=0; j<100; j++);
    }
//    printf("Thread %d: and now a delay of %d clock ticks\n", message, delay);
    cyg_thread_delay(delay);
    c = tmu_task_tmr_get_count();
    tmu_task_tmr_set_irqen(0); // prevents TMU irq while in printf
    printf("Thread %d: used %7u TMU cycles\n", arg->message, c);
    tmu_task_tmr_set_irqen(1);
  }
}



void show_threads(void)
{
  cyg_handle_t thread = 0;
  cyg_uint16 id = 0;
  cyg_thread_info info;

  printf("  --> Current thread is handle: 0x%x\n", cyg_thread_self());

  while( cyg_thread_get_next( &thread, &id ) ) {

    if( !cyg_thread_get_info( thread, id, &info ) )
      break;

    printf("ID: %04x name: %15s pri: %2d state: %d handle: 0x%x\n",
           info.id, info.name?info.name:"----", info.set_pri, info.state, info.handle);
  }
}

